//************************************************************
//  University of Patras 
//	Department of Electrical Engineering & Computer Technology
//	 
//	Course: Parallel Programming in AI 
//	Week 01: Code Optimization
//
//  Description: Arithmetic operations of Large two dimensional Matrices
//  	addmatOrg05.c: Using pointers to accelerate the program execution
//						and the estimation of the matrix product is derived
//						through the temporal variable s
//                      Local variables in CPU registers
//	Author: Evangelos Dermatas
//	v0.1
/*
*********************************************************************
gcc -v:	gcc version 7.4.0 (Ubuntu 7.4.0-1ubuntu1~18.04.1)
OS: 	Linux Ubuntu 18.04
CPU:	8-core, Intel(R) Core(TM) i7-2630QM CPU @ 2.00GHz, 3991 BogoMips
*********************************************************************
Compiling:	gcc addmatOrg05.c -o addmatOrg05
Executing:	time ./addmatOrg05
Output in my Laptop:
innerProdMat  207115863

real	0m8.309s
user	0m8.301s
sys		0m0.008s
*********************************************************************
Compiling:	gcc addmatOrg05.c -o addmatOrg05 -O2
Executing:	time ./addmatOrg05
Output in my Laptop:

innerProdMat  207115863

real	0m8.930s
user	0m8.926s
sys		0m0.004s
*********************************************************************
Compiling:	gcc addmatOrg05.c -o addmatOrg04 -O4
Executing:	time ./addmatOrg05
Output in my Laptop:

innerProdMat  207115863

real	0m9.835s
user	0m9.827s
sys		0m0.008s
*********************************************************************
Compiler provided by intel
icc -v: icc version 19.0.5.281 (gcc version 7.4.0 compatibility)
Compiling:	icc addmatOrg05.c -o addmatOrg05
Executing:	time ./addmatOrg05

innerProdMat  207115863

real	0m8.960s
user	0m8.951s
sys		0m0.008s

*/
//************************************************************



#include <stdio.h>
#include <stdlib.h>
//#include <omp.h>

// Matrices Dimension
#define N	1000

// **************************************************************
int A[N][N] ;
int B[N][N] ;
int C[N][N] ;

// ***************************************************************************
//    Set random values to the elements of a two dimensional matrix a[N][N] 
// 		in the range [-50,49]
// ***************************************************************************
void setValMat( int a[N][N] ) 
{
	register int i ;
	register int *p ;

	for( i = 0, p = (int *) a ; i < N*N ; i++ )
		    *p++ = (rand() % 100) - 50 ; 
}

// ***************************************************************************
//    Add matrix a to b into c
// ***************************************************************************
void addMat( int a[N][N], int b[N][N], int c[N][N] ) 
{
	register int i, *pa, *pb, *pc ;

	for( i = 0, pa = (int *) a, pb = (int *) b, pc=(int *) c ; i < N*N ; i++ )
	    	*pc++ = *pa++ + *pb++ ; 
}

// ***************************************************************************
//    Compute the inner product of a and b
// ***************************************************************************
int innerProdMat( int a[N][N], int b[N][N] ) 
{
	register int i, *pa, *pb, sum = 0 ;

	for( i = 0, pa = (int *) a, pb = (int *) b ; i < N*N ; i++ )
		    sum += *pa++  *  *pb++ ; 
	return sum ;
}


// ***************************************************************************
//    Compute in c the product of a and b
// ***************************************************************************
void mulMat( int a[N][N], int b[N][N], int c[N][N] ) 
{
	register int i, j, k, *pa, *pb, *pc, s ;

   	for( i = 0, pc = (int *) c ; i < N ; i++ )
		for( j = 0 ; j < N ; j++, pc++ ) {
	        for( k = s = 0, *pc = 0, pa=&a[i][0], pb=&b[0][j] ; k < N ; k++, pb+=N )
		        s += *pa++ * *pb ; 
		*pc = s ;
		}
}


// **************************************************************
int main(void) 
{
register double t1, t2, ttot = 0, t ;
register int res ;


//	t1 = omp_get_wtime() ;
	setValMat( A ) ;
//	t2 = omp_get_wtime() ;
//	ttot = t = t2-t1 ;
//	printf( "SetVal A      %lf\n", t ) ;

//	t1 = omp_get_wtime() ;
	setValMat( B ) ;
//	t2 = omp_get_wtime() ;
//	ttot += t = t2-t1 ;
//	printf( "SetVal B      %lf\n", t ) ;

//	t1 = omp_get_wtime() ;
	addMat( A, B, C ) ; 
//	t2 = omp_get_wtime() ;
//	ttot += t = t2-t1 ;
//	printf( "addMat        %lf\n", t ) ;

//	t1 = omp_get_wtime() ;
	mulMat( C, B, A ) ; 
//	t2 = omp_get_wtime() ;
//	ttot += t = t2-t1 ;
//	printf( "MulMat        %lf\n", t) ;

//	t1 = omp_get_wtime() ;
	res = innerProdMat( A, B ) ; 
//	t2 = omp_get_wtime() ;
//	ttot += t = t2-t1 ;
//	printf( "innerProdMat  %lf %d\n", t, res ) ;
	printf( "innerProdMat  %d\n", res ) ;

//	printf( "Total Time:   %lf\n", ttot) ;

	return 0 ;
}

